#########################################################################
# mmucl_gui.tcl - library for using Mmucl in Tk
#
# Copyright (C) 1997-1999 Mark Patton
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#########################################################################

namespace eval mmucl {
    array set Mmucl {
	cfg_echo        1
	cfg_echo_color  yellow
	cfg_keep_line   1
	cfg_hist_min    3
	cfg_hist_keep   20
    }
    
    array set Mmucl {
	history         {""}
	hist_loc        0
	hist_cur        ""
    }
    lappend Mmucl(dump) key
    set Mmucl(lib_dir) $env(MMUCL_DIR)
}

proc mmucl::gui_init {} {
    variable Mmucl
    global tcl_platform

    foreach img [glob [file join $Mmucl(lib_dir) images *.gif]] {
	image create photo [file tail $img] -file $img
    }
    
    # icon (unix only)
    if {[string compare $tcl_platform(platform) "unix"] == 0} {
	toplevel .icon
	label .icon.l -image mmucl.gif -bd 0
	pack .icon.l
	wm iconwindow . .icon
    }
    
    set conf [file join $Mmucl(lib_dir) Tk.conf]
    if {[file exists $conf] && [catch {option readfile $conf} error]} {
	puts stderr "loading $conf: $error"
    }
    
    set conf [file join $Mmucl(rc_dir) Tk.conf]
    if {[file exists $conf] && [catch {option readfile $conf} error]} {
	puts stderr "loading $conf: $error"
    }

    wm protocol . WM_DELETE_WINDOW mmucl::MCexit

    return
}

proc mmucl::user_entry {w} {
    variable Mmucl
    
    set Mmucl(entry) $w
    
    entry $w
    bind $w <Key-Return> {mmucl::meta_parse %W}
    bind $w <Key-Up> {mmucl::hist_scroll_up %W}
    bind $w <Key-Down> {mmucl::hist_scroll_down %W}
    
    bind $w <Control-Shift-Escape> {
	if {[llength [mmucl alias debug]]} {
	    mmucl alias debug {}
	} else {
	    mmucl alias debug uplevel #0
	}
    }
    
    # disable tab focus switching
    bind $w <Tab> break
    
    # MCkey binds scripts to mmucl
    bindtags $w [concat mmucl [bindtags $w]]
    
    return $w 
}

bind Entry <Control-u> {%W delete 0 insert}

proc mmucl::toolbar {w} {
    variable Mmucl

    frame $w -class Toolbar
    
    button $w.con -command mmucl::tk_connect -image connect.gif 
    balloon_add $w.con "Connect to mud"
    button $w.load -image load.gif -command {mmucl::tk_source}
    balloon_add $w.load "Load a file"
    button $w.send -image send.gif -command {mmucl::tk_textin}
    balloon_add $w.send "Send a file"
    button $w.save -image save.gif -command {mmucl::tk_dump}
    balloon_add $w.save "Save state"
    button $w.char -image char.gif -command {mmucl::tk_edit char}
    balloon_add $w.char "Edit characters"
    button $w.alias -image alias.gif -command {mmucl::tk_edit alias}
    balloon_add $w.alias "Edit aliases"
    button $w.action -image action.gif -command {mmucl::tk_edit action}
    balloon_add $w.action "Edit actions"
    button $w.sub -image sub.gif -command {mmucl::tk_edit sub}
    balloon_add $w.sub "Edit subs"
    button $w.key -image bind.gif -command {mmucl::tk_edit key}
    balloon_add $w.key "Edit keys"

    pack $w.con $w.load $w.send $w.save $w.char $w.alias $w.action $w.sub \
	$w.key -side left 

    return $w
}

proc mmucl::option_menu {w} {
    menu $w

    $w add checkbutton -label Reconnect -variable mmucl::Mmucl(cfg_reconnect)
    $w add checkbutton -label "Keep line" -variable mmucl::Mmucl(cfg_keep_line)
    $w add checkbutton -label Actions -variable mmucl::Mmucl(cfg_actions)
    $w add checkbutton -label Subs -variable mmucl::Mmucl(cfg_subs)
    $w add checkbutton -label "Strip ANSI" \
	    -variable mmucl::Mmucl(cfg_strip_ansi)
    $w add checkbutton -label "Broken EOL" \
	    -variable mmucl::Mmucl(cfg_broken_eol)
    $w add checkbutton -label "Local echo" -variable mmucl::Mmucl(cfg_echo)
    $w add cascade -label "Echo color" -menu $w.colors
    
    menu $w.colors
    foreach color {red green yellow blue magenta cyan grey} {
	$w.colors add radiobutton -label $color -value $color \
	    -variable mmucl::Mmucl(cfg_echo_color)
    }

    return $w
}

proc mmucl::meta_parse {w} {
    variable Mmucl

    set input [$w get]

    if {$Mmucl(cfg_echo)} {
	MCecho [MCcolor $input $Mmucl(cfg_echo_color)]
    }
    
    if {[catch {MCparse $input} val]} {
	report error $val
    } elseif {[string length $val]} {
	MCecho $val
    }

    if {[string length $input] >= $Mmucl(cfg_hist_min)} {
	set Mmucl(history) [lreplace $Mmucl(history) 0 \
		[expr {[llength $Mmucl(history)] - $Mmucl(cfg_hist_keep)}]]
	lappend Mmucl(history) $input
	set Mmucl(hist_loc) [llength $Mmucl(history)]
    }

    if {$Mmucl(cfg_keep_line)} {
	$w selection range 0 end
    } else {
	$w delete 0 end
    }

    return
}

proc mmucl::hist_scroll_up {w} {
    variable Mmucl

    if {$Mmucl(hist_loc)} {
	if {$Mmucl(hist_loc) == [llength $Mmucl(history)]} {
	    set Mmucl(hist_cur) [$w get]
	}	
	$w delete 0 end
	$w insert 0 [lindex $Mmucl(history) [incr Mmucl(hist_loc) -1]]
    }
    
    return
}

proc mmucl::hist_scroll_down {w} {
    variable Mmucl

    if {[incr Mmucl(hist_loc)] > [llength $Mmucl(history)]} {
	incr Mmucl(hist_loc) -1
    } else {
	$w delete 0 end
	if {$Mmucl(hist_loc) == [llength $Mmucl(history)]} {
	    $w insert 0 $Mmucl(hist_cur)
	} else {
	    $w insert 0 [lindex $Mmucl(history) $Mmucl(hist_loc)]
	}
    }
    
    return
}

proc mmucl::tk_script {script descript} {
    global errorInfo

    set got_err 0
    if {[catch $script error]} {
	set act [tk_dialog .err "Error: $descript" $error error 0 Ok\
		"Print stack trace"]
	
	if {$act == 1} {
	    MCecho "Stack trace:"
	    MCecho $errorInfo
	}

	return 0
    }

    return 1
}

# ugly
proc mmucl::tk_connect {} {
    variable Mmucl
    
    if {[winfo exists .con]} {
	raise .con
	return
    }
    
    toplevel .con
    wm title .con Connect

    frame .con.f -relief groove -bd 2
    grid [label .con.f.hostl -text Host:]
    grid [entry .con.f.host] -row 0 -column 1
    grid [label .con.f.portl -text Port:] 
    grid [entry .con.f.port] -row 1 -column 1
    
    .con.f.host insert end $Mmucl(host)
    .con.f.port insert end $Mmucl(port)

    if {![info exists Mmucl(tmpchar)]} {
	set Mmucl(tmpchar) ""
    }

    # if characters defined, make a character selection area
    if {[llength [set chars [MCchar names]]]} {
	grid [label .con.f.lchar -text Char:]

        set m [eval tk_optionMenu .con.f.char mmucl::Mmucl(tmpchar) $chars]
	$m add separator
	$m add command -label None -command {set mmucl::Mmucl(tmpchar) ""}
	set Mmucl(tmpchar) ""

	grid .con.f.char -row 2 -column 1
	
	# hack to update host and port on char selection.
	bind $m <ButtonRelease> {
	    .con.f.host delete 0 end
	    .con.f.port delete 0 end
	    
	    after 10 {
		if {[string length $mmucl::Mmucl(tmpchar)]} {
		    .con.f.host insert 0 \
			[lindex [mmucl::MCchar set $mmucl::Mmucl(tmpchar)] 0]
		    .con.f.port insert 0 \
			[lindex [mmucl::MCchar set $mmucl::Mmucl(tmpchar)] 1]
		}
	    }
	}
    }
    
    frame .con.b
    pack [button .con.b.ok -text Connect \
	    -command mmucl::tk_connect_do] -side left
    pack [button .con.b.quit -text Cancel -command {destroy .con}] -side left
  
    pack .con.f .con.b
    wm geometry .con +[winfo pointerx .]+[winfo pointery .]
    wm resizable .con 0 0

    return
}

proc mmucl::tk_connect_do {} {
    variable Mmucl

    if {[string length $Mmucl(tmpchar)]} {
	set res [tk_script [list MCchar load $Mmucl(tmpchar)] "Loading char"]
    } else {
	set res [tk_script [list MCconnect [.con.f.host get]\
		[.con.f.port get]] "Connecting to host"]
    }
    
    if {$res} {
	destroy .con
    }

    return
}

proc mmucl::tk_edit {proc} {
    set list 0
    switch -exact -- $proc {
	action {
	    set labels {Pattern: Script:}
	    set title "Edit Actions"
	} alias {
	    set labels {Name: Script:}
	    set title "Edit Aliases"
	} sub {
	    set labels {Pattern: Subspec:}
	    set title "Edit Subs"
	} key {
	    set labels {Event: Script:}
	    set title "Edit Keys"
	} char {
	    set labels {Name: Host: Port: Login:}
	    set title "Edit Characters"
	    set list 1
	}
    }

    set w .$proc

    if {[winfo exists $w]} {
	raise $w
	return
    }
  
    toplevel $w
    wm title $w $title

    form $w.form $labels MC$proc $list
    pack $w.form -side top -pady 3
 
    frame $w.b
    pack [button $w.b.help -text Help -command [list mmucl::MChelp $proc]]\
	    [button $w.b.exit -text Close -command [list destroy $w]] \
	    -side left
    pack $w.b -side bottom
    
    wm geometry $w +[winfo pointerx .]+[winfo pointery .]
    wm resizable $w 0 0

    return
}

proc mmucl::MCkey {args} {
    set syntax {
	 set      {{+ key} {? script}}
	 names    {{? pattern *}}
	 print    {{? pattern *}}
	 delete   {{- exact} {+ pattern}}
     }
     
     foreach name [bind mmucl] {
	 set keys($name) [lindex [bind mmucl $name] 2]
     }
     
     switch [check key $syntax $args 1] {
	 set {
	     if {[info exists arg(script)]} {
		 bind mmucl $arg(key) \
			 [list mmucl::key_exec $arg(key) $arg(script)]
	     } elseif {![info exist keys($arg(key))]} {
		 error "no such key"
	     } else {
		 return $keys($arg(key))
	     }
	 } names {
	     return [array names keys $arg(pattern)]
	 } delete {
	     if {[info exists arg(-exact)]} {
		 bind mmucl $arg(pattern) {}
	     } else {
		 foreach key [array names keys $arg(pattern)] {
		     bind mmucl $key {}
		 }
	     }
	 } print {
	     foreach key [array names keys $arg(pattern)] {
		 MCecho "$key bound to {$keys($key)}"
	     }
	 }
     }
    
    return
}

# relies on error text to propogate break

proc mmucl::key_exec {key script} {
    if {[catch {mmucl eval $script} error]} {
	if {[string match *break* $error]} {
	    return -code break
	}

	report error "key $key: $error"
    }
    
    return
}

proc mmucl::MChelp {{subject ""}} {
    variable Mmucl
    global tki tkiEmbed env tcl_platform

    if {![info exists tki]} {
	set tkiEmbed 1

	# find tkinfo somewhere in the users path
	if {[string compare $tcl_platform(platform) "unix"] == 0} {
	    set path [split $env(PATH) :]
	} elseif {[string compare $tcl_platform(platform) "windows"] == 0} {
	    set path [split $env(PATH) \;]
	} else {
	    set path [pwd]
	}

	foreach dir $path {
	    if {[file exists [file join $dir tkinfo]]} {
		uplevel #0 [list source [file join $dir tkinfo]]
		break
	    }
	}
	
	if {![info exists tki]} {
	    error "tkinfo not in path"
	}
    }
    
    tkiWinShow (mmucl)$subject
    
    return
}

# Modify the command line.

proc mmucl::MCcline {args} {
    variable Mmucl

    set w $Mmucl(entry)
    set syntax {
	delete {{? first 0} {? last end}}
	get {{? first 0} {? last end}} 
	insert {{+ first} {+ str}}
	history {}
    }
    
    switch -exact [check cline $syntax $args 1] {
        delete {
	    $w delete $arg(first) $arg(last)
	} get {
	    return [string range [$w get] \
		    [$w index $arg(first)] [$w index $arg(last)]]
	} insert {
	    $w insert $arg(first) $arg(str)
	} history {
	    return $Mmucl(history)
	}
    }

    return
}

proc mmucl::tk_source {} {
    set file [tk_getOpenFile -title "Load a file"]
    
    if {[string length $file]} {
	tk_script [list mmucl invokehidden source $file] "Loading script"
    }
    
    return
}

proc mmucl::tk_textin {} {
    set file [tk_getOpenFile -title "Send a file"]
    
    if {[string length $file]} {
	tk_script [list MCtextin $file] "Sending file"
    }
    
    return
}

proc mmucl::tk_dump {} {
    set file [tk_getSaveFile -title "Save state to file"]
   
    if {[string length $file]} {
	tk_script [list MCdump -all -- $file] "Saving state"
    }

    return
}

proc mmucl::tk_exit {} {
    set s [tk_messageBox -type yesno -message "Really quit?" -icon question]
    
    if {[string compare $s "yes"] == 0} {
	MCexit
    }
    
    return
}

proc mmucl::form {w labels proc list} {
    frame $w -class Form
    frame $w.sel
    pack [listbox $w.sel.list -yscroll [list $w.sel.yscroll set]] -side left
    pack [scrollbar $w.sel.yscroll -command [list $w.sel.list yview]] \
	      -side left -fill y

    frame $w.edit
    frame $w.edit.buts
    pack [button $w.edit.buts.add -text Add \
	      -command [list mmucl::form_add $w $proc $list]] -side left
    pack [button $w.edit.buts.delete -text Delete \
	      -command [list mmucl::form_delete $w $proc]] -side left
    frame $w.edit.f
    
    set i 0
    foreach label $labels {
	grid [label $w.edit.f.l$i -text $label] -sticky w -column 0
	grid [entry $w.edit.f.e$i] -row $i -sticky ew -column 1
	incr i
    }
    
    pack $w.edit.f $w.edit.buts -side top
    pack $w.sel $w.edit -side left
   
    bind $w.edit.f.e[incr i -1] <Key-Return> \
	    [list mmucl::form_add $w $proc $list]
    bind $w.sel.list <ButtonRelease-1> \
	[list mmucl::form_show_selection $w $proc $list]

    form_listbox_update $w $proc

    return $w
}

proc mmucl::form_show_selection {w proc list} {
    set i [$w.sel.list curselection]

    if {[string length $i]} {
	set i [$w.sel.list get $i]

	if {$list} {
	    form_set_entries $w [concat [list $i] [$proc set $i]]
	} else {
	    form_set_entries $w [list $i [$proc set $i]]
	}
    }

    return
}

proc mmucl::form_add {w proc list} {
    set e [form_get_entries $w]
    
    if {$list} {
	set script [list $proc set [lindex $e 0] [lrange $e 1 end]] 
    } else {
	set script  [concat [list $proc] set $e]
    }

    # assumes MCname procs...
    tk_script $script "Adding [string range $proc 2 end]"
    form_listbox_update $w $proc

    return
}

proc mmucl::form_delete {w proc} {
    set i [$w.sel.list curselection]
    
    if {[string length $i]} {
	set i [$w.sel.list get $i]
	$proc delete -exact -- $i
	form_listbox_update $w $proc
    }
    
    return
}

proc mmucl::form_listbox_update {w proc} {
    form_set_entries $w {}

    $w.sel.list delete 0 end
    foreach i [lsort [$proc names]] {
        $w.sel.list insert end $i
    }

    return
}

proc mmucl::form_get_entries {w} {
    set num_entries [expr {[llength [winfo children $w.edit.f]]/2}]

    for {set i 0} {$i < $num_entries} {incr i} {
	lappend list [$w.edit.f.e$i get]
    }
    return $list
}

proc mmucl::form_set_entries {w list} {
    set num_entries [expr {[llength [winfo children $w.edit.f]] / 2}]

    for {set i 0} {$i < $num_entries} {incr i} {
 	$w.edit.f.e$i delete 0 end
 	$w.edit.f.e$i insert end [lindex $list $i]
    }

    return
}